require "threaded"

# RSVP middleware to transform 'RSVP' requests into corresponding
# contacts in your Google Contacts via Google People API.
class Oxy::RSVP
  # The set of allowed fields. Requests that do not have fields
  # present in this list will not be eligible to be enqueued.
  ELIGIBLE_FORMS_FIELDS = ['email_address', 'first_name', 'last_name']
  # Email validation regex
  EMAIL_REGEX_VALIDATOR = /\A([^@\s]+)@((?:[-a-z0-9]+\.)+[a-z]{2,})\z/i
  # First and last name allowed string length
  ALLOWED_STRING_LENGTH = 60

  # ctor.
  def initialize(app, logger = $stderr)
    @app = app
    @logger = logger
    # middleware to deflect abusive clients, 10/90 rps
    @deflect = Rack::Deflect.new(app, :log => logger, :request_threshold => 10, :interval => 90, :log_format => "[Rack::Deflect]: (%s) ~> %s")
  end

  # Middleware's entry point
  def call(env)
    # instantiate the request object
    req = Rack::Request.new(env)
    if req.path == "/rsvp" && req.post?
      # first ask deflect layer for permission to process the request
      resp = @deflect.call(env)
      return resp if forbidden(resp)
      # next ask validation layer for permission to process the request
      # ...
      # enqueue background processing for valid submissions only
      Threaded.enqueue(Subscribe, req.POST, @logger) if valid_form(req.POST)
      # redirect anyways
      [302, { "Location" => "/thank-you" }, []]
    else
      @app.call(env)
    end
  end

private
  # simple helper to find out whether the request is forbidden or not
  def forbidden(resp)
    resp[0] == 403
  end

  def validate_email(value)
    # match or no match?
    EMAIL_REGEX_VALIDATOR.match?(value)
  end

  def validate_length(value)
    value.nil? || value.empty? || value.length <= ALLOWED_STRING_LENGTH
  end

  # only request with eligible and valid form fields are allowed
  def valid_form(form)
    # validate fields eligibility
    is_valid = form.all? { |key, _| ELIGIBLE_FORMS_FIELDS.include?(key) }
    # validate email address
    is_valid = false unless validate_email(form["email_address"]) && is_valid
    # validate first name
    is_valid = false unless validate_length(form["first_name"]) && is_valid
    # validate last name
    is_valid = false unless validate_length(form["last_name"]) && is_valid
    # form is not valid
    unless is_valid
      @logger.write("[RSVP]: Received an invalid form ~> #{form.inspect}\n")
      return false
    end
    # form is valid
    return true
  end
end